/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Die Sep 16 10:25:46 EDT 2003
    copyright            : (C) 2003 by VWP
    email                : virtual_worlds@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
* PURPOSE:
* The LoadController is a program which periodically checks the system load.
* If the load becomes too high, all processes are checked and the ones which
* consume too much CPU time AND are owned by a user which has to be
* controlled, are killed. Additionally an information email is send.
* The option to kill only processes of specified users avoids heavy trouble
* or system instabilities. Of course the user root should never be
* configured as a user which has to be controlled
*/


 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <dirent.h>
#include <pwd.h>
#include <sys/types.h>

typedef unsigned char boolean;
#define false 0
#define true  1

#define LANGUAGE_DE 1
#define LANGUAGE_EN 2



// structure to store the retrieved process data into
struct procData
   {
   struct procData *Next;
   int              PID,UID;  // ID of process, ID of the owner
   long             utime;
   char             PName[300];
   float            loadPerc; // the actual load in percent
   struct userList *OwnerUL;  // pointer to the owners information
   };

// a structure which contais the configuration data (default or read from /etc/loadctrl.config
struct configData
   {
   int              checkInterval; // interval in seconds to check the cpu load
   float            loadLimit;     // the maximum system load which is allowed
   int              usageLimit;    // the maximum cpu usage which is allowed
   char             sendmail[300]; // path to sendmail
   int              language;      // specifies the language of the email
   };

// contains the list of watched users and their email addresses as specified in /etc/loadctrl.config
struct userList
   {
   struct userList *Next;
   char             UName[300],EMail[300];
   int              UID;
   };



struct procData   DummyPD={NULL,-1,-1,-1,"",0.0,NULL};
struct procData  *FirstPD=&DummyPD;
struct configData UseConfig={60,5.0,25,"/usr/sbin/sendmail",LANGUAGE_DE};
struct userList   DummyUL={NULL,"","",-1};
struct userList  *FirstUL=&DummyUL;
boolean           operating=false,leaveDaemon=false;



/***************************************************************************************************
* This function checks if a users ID can be found in the list of users specified in the
* configuration file.
* Input:  CheckPD - a process data structure which contains the user ID of the process a owner is
*         searched for
* Return: true if the owner of the process is specified in the user list, fasle otherwise
* Side Effects: In case the User was found in the user list, CheckPD->OwnerUL will be set to point
*         to the owners user list entry. Thats necessary to send the information email because the
*         email address can be found in the user data list only
*/
boolean isInUserList(struct procData *CheckPD)
{
struct userList *CheckUL=FirstUL->Next;

while (CheckUL)
   {
   if (CheckUL->UID==CheckPD->UID)
      {
      CheckPD->OwnerUL=CheckUL;
      return true;
      }
   CheckUL=CheckUL->Next;
   }
return false;
}



/***************************************************************************************************
* Kills a specified process and sends an information email to the owner.
* Input:  KillPD - process data of the process which has to be killed
*         actLoad - the actual server load, this value is used for the information email only
*/
void killProc(struct procData *KillPD,float actLoad)
{
FILE *FHandle;
char mailStr[300];

kill(KillPD->PID,SIGINT);
kill(KillPD->PID,SIGQUIT);
kill(KillPD->PID,SIGTERM);
kill(KillPD->PID,SIGKILL);

sprintf(mailStr,"%s %s\n",UseConfig.sendmail,KillPD->OwnerUL->EMail);
FHandle=popen(mailStr,"w");
if (FHandle)
   {
   fprintf(FHandle,"To: %s\n",KillPD->OwnerUL->EMail);
   if (UseConfig.language==LANGUAGE_DE)
      {
      fprintf(FHandle,"Subject: Prozess gettet!\nACHTUNG!\n\n");
      fprintf(FHandle,"Ihr Proze %s wurde beendet weil er zu viel CPU-Zeit verbraucht (%.2f %%) \
und eine Serverlast von %.2f erzeugt hat.\n\
Bitte berprfen Sie sofort Ihre Konfiguration um weitere schwerwiegende Strungen des Systems zu vermeiden.\n.\n",
              KillPD->PName,KillPD->loadPerc,actLoad);
      }
   else
      {
      fprintf(FHandle,"Subject: Process killed!\nATTENTION!\n\n");
      fprintf(FHandle,"Your process %s was killed because it consumed too much CPU time (%.2f %%) and caused a too high server load of %.2f\n\
Please check your configuration immediately to avoid further negative influence on the system.\n.\n",
              KillPD->PName,KillPD->loadPerc,actLoad);
      }
   pclose(FHandle);
   }
else
   {
   printf("ERROR: Could not execute mailprogram \"%s\"!\n",UseConfig.sendmail);
   printf("Killing %s (%d) from %s / %s (%f %% CPU usage, %f load)\n",KillPD->PName,KillPD->PID,KillPD->OwnerUL->UName,KillPD->OwnerUL->EMail,KillPD->loadPerc,actLoad);
   }
}



/***************************************************************************************************
* This function contains the main functionality of the process. Here information of all processes
* are read from /proc. If they use a system time >0 (what means they are running), a process data
* structure is created for them. In a second step this list is checked and for processes which
* require a CPU time of more than 1% the ID of the owner is fetched. If a process exceeds the CPU
* usage limit specified in UseConfig.usageLimit AND if it is owned by a user whic hhas to be
* controlled by this daemon, the process will be killed.
* If no process could be killed afterwards the process with the highest CPU usage is killed if
* it consumes more than 1% CPU AND is owned by one of the checked users.
* Input:  actLoad - specifies the server load; this value is only forwarded to the killProc()
*         function
* Return: true if the function worked properly, false otherwise
*/
boolean killAmokProcess(float actLoad)
{
#define MAX_COMMLENGTH 300

struct procData *NewPD,*MaxLoadPD=NULL;
DIR             *DHandle;
FILE            *FHandle;
struct dirent   *procDir;
char             FName[300];
long             totalUTime=0;
char             comm[MAX_COMMLENGTH];
char             state;
int              pid,ppid,pgrp,session,tty,tpgid,flags,minflt,cminflt,majflt,cmajflt,utime,stime,
                 cutime,cstime,counter,priority,timeout,itrealvalue,starttime,vsize,rss,rlim,
                 startcode,endcode,startstack,kstkesp,kstkeip,signal,blocked,sigignore,sigcatch,
                 wchan;
float            maxLoadPerc=0,actLoadPerc;
int              tmp1,tmp2,tmp3;
boolean          procKilled=false;
                 
DHandle=opendir("/proc");
if (DHandle)
   {
   do
      {
      procDir=readdir(DHandle);
      if (procDir)
         {
         if (isdigit(procDir->d_name[0]))
            {
            sprintf(FName,"/proc/%s/stat",procDir->d_name);
            FHandle=fopen(FName,"r");
            if (FHandle)
               {
               fscanf(FHandle,"%d %s %c %d %d %d %d %d %u %u %u %u %u \
                               %d %d %d %d \
                               %d %d %u %u %d %u %u %u %u %u %u %u %u %d %d %d %d %u",
                               &pid,comm,&state,&ppid,&pgrp,&session,&tty,&tpgid,&flags,&minflt,&cminflt,&majflt,&cmajflt,
                               &utime,&stime,&cutime,&cstime,
                               &counter,&priority,&timeout,&itrealvalue,&starttime,&vsize,&rss,&rlim,&startcode,&endcode,&startstack,&kstkesp,&kstkeip,&signal,&blocked,&sigignore,&sigcatch,&wchan);
               fclose(FHandle);
               totalUTime+=utime;
               if (utime>0) // thread is running so we store its data
                  {
                  NewPD=(struct procData*)malloc(sizeof(struct procData));
                  if (NewPD)
                     {
                     memset(NewPD,0,sizeof(struct procData));
                     NewPD->Next=FirstPD->Next; FirstPD->Next=NewPD;
                     NewPD->PID=pid;
                     NewPD->UID=-1;
                     NewPD->utime=utime;
                     strcpy(NewPD->PName,comm);
                     }
                  else printf("ERROR: Can't allocate memory for process data list!");
                  }
               }
            else printf("ERROR: cant access information in %s!\n",FName);
            }
         }
      }
   while (procDir);
   closedir(DHandle);
   while (FirstPD->Next)
      {
      NewPD=FirstPD->Next;
      actLoadPerc=NewPD->utime*100.0/totalUTime;
      if (actLoadPerc>1) // everything below 1 % load is not interesting for us
         {
         sprintf(FName,"/proc/%d/status",NewPD->PID);
         FHandle=fopen(FName,"r");
         if (FHandle)
            {
            while (fgets(comm,MAX_COMMLENGTH,FHandle)>0)
               {
               if (strstr(comm,"Uid:")==comm)
                  {
                  sscanf(comm,"Uid:\t%d\t%d\t%d\t%d",&NewPD->UID,&tmp1,&tmp2,&tmp3);
                  break;
                  }
               }
            fclose(FHandle);
            NewPD->loadPerc=actLoadPerc;
            if ((actLoadPerc>UseConfig.usageLimit) && (isInUserList(NewPD)))
               {
               procKilled=true;
               killProc(NewPD,actLoad);
               }
            if ((actLoadPerc>maxLoadPerc) && (isInUserList(NewPD)))
               {
               maxLoadPerc=actLoadPerc;
               MaxLoadPD=NewPD;
               }
            }
         }
      FirstPD->Next=FirstPD->Next->Next;
      free(NewPD);
      }
   if ((MaxLoadPD) && (!procKilled)) killProc(MaxLoadPD,actLoad);
   }
else printf("ERROR: could not access /proc!\n");
return false;
}



/***************************************************************************************************
* Releases all memory used by the daemon. At the moment this influences only the user list.
* Side Effects: No user data list is available afterwards
*/
void freeMem(void)
{
struct userList *RemUL;

while (FirstUL->Next)
   {
   RemUL=FirstUL->Next;
   FirstUL->Next=FirstUL->Next->Next;
   free(RemUL);
   }
}



/***************************************************************************************************
* Reads the program configuration file /etc/loadctrl.config and stores the read data in UseConfig
* and inside a user data list.
* Return:  True if the data configuration could be read and stored successfully and if there were
*          no fatal errors inside of it, false otherwise
* Side Effects: A user data list will be created starting with FirstUL->Next
*/
boolean loadConfig(void)
{
#define MAX_CONFIGLENGTH 300
#define CONFIG_FILE      "/etc/loadctrl.config"

FILE            *FHandle;
char             configStr[MAX_CONFIGLENGTH];
struct passwd   *UserInfo;
struct userList *NewUL;

FHandle=fopen(CONFIG_FILE,"r");
if (FHandle)
   {
   while (fgets(configStr,MAX_CONFIGLENGTH,FHandle)>0)
      {
      if (strstr(configStr,"CHECKINTERVAL")==configStr)
       sscanf(configStr,"CHECKINTERVAL=%d",&UseConfig.checkInterval);
       
      else if (strstr(configStr,"LOADLIMIT")==configStr)
       sscanf(configStr,"LOADLIMIT=%f",&UseConfig.loadLimit);
       
      else if (strstr(configStr,"USAGELIMIT")==configStr)
       sscanf(configStr,"USAGELIMIT=%d",&UseConfig.usageLimit);
       
      else if (strstr(configStr,"SENDMAIL")==configStr)
       sscanf(configStr,"SENDMAIL=%s",UseConfig.sendmail);

      else if (strstr(configStr,"LANGUAGE=de")==configStr) UseConfig.language=LANGUAGE_DE;
      else if (strstr(configStr,"LANGUAGE=")==configStr)   UseConfig.language=LANGUAGE_EN;
       
      else if (strstr(configStr,"CHECKUSER")==configStr)
         {
         NewUL=(struct userList*)malloc(sizeof(struct userList));
         if (NewUL)
            {
            sscanf(configStr,"CHECKUSER %s %s",NewUL->UName,NewUL->EMail);           
            UserInfo=getpwnam(NewUL->UName);
            if (UserInfo)
               {
               NewUL->UID=UserInfo->pw_uid;
               NewUL->Next=FirstUL->Next;
               FirstUL->Next=NewUL;
               }
            else
               {
               printf("ERROR: Invalid username %s in directive\"%s\"!\n",NewUL->UName,configStr);
               free(NewUL);
               }
            }
         else
            {
            fclose(FHandle);
            printf("ERROR: Not enough memory to read configuration data!\n");
            return false;
            }
         }
      } 
   fclose(FHandle);
   if (!FirstUL->Next)
      {
      printf("ERROR: No valid users specified which have to be controlled!\n");
      return false;
      }
   return true;
   }
else
   {
   printf("ERROR: Could not find configuratiuon file \"%s\"!\n",CONFIG_FILE);    
   return false;
   }
}



/***************************************************************************************************
* This function handles incoming signals and breaks up the program if necessary.
* Input:  signal - the received signal, this parameter is handed over by the system
* Side effects: the global variable leaveDaemon will be set to true if a signal was received which
*         causes a program stop
*/
void signalHandler(int signal)
{
if ((signal==SIGINT) || (signal==SIGQUIT) ||(signal==SIGTERM))
   {
   if (!operating)
      {
      leaveDaemon=true;
      freeMem();
      }
   else leaveDaemon=true;
   }
}



int main(int argc, char *argv[])
{
FILE *FHandle;
float load1,load2,load3;
char  temp1[200],temp2[200];

signal(SIGINT,signalHandler);
signal(SIGQUIT,signalHandler);
signal(SIGTERM,signalHandler);
if (!loadConfig())
   {
   freeMem();
   return -1;
   }
daemon(0,0);
while (!leaveDaemon)
   {
   FHandle=fopen("/proc/loadavg","r");
   if (FHandle)
      {
      fscanf(FHandle,"%f %f %f %s %s",&load1,&load2,&load3,temp1,temp2);
      // 1, 5, 15 min
      operating=true;
      if ((UseConfig.checkInterval<=60) && (load1>UseConfig.loadLimit)) killAmokProcess(load1);
      else if ((UseConfig.checkInterval>600) && (load3>UseConfig.loadLimit)) killAmokProcess(load3);
      else if (load2>UseConfig.loadLimit) killAmokProcess(load2);
      operating=false;
      fclose(FHandle);
      }
   else
      {
      printf("Error: /proc must be mounted\n  To mount /proc at boot you need an /etc/fstab line like:\n      /proc   /proc   proc    defaults\n  In the meantime, mount /proc /proc -t proc");
      break;
      }
   if (!leaveDaemon) sleep(UseConfig.checkInterval);
   }
freeMem();
return 0;
}


